@@ -25,6 +25,8 @@
 
 """Command line parsing and processing for twyt"""
 
+import sys
+import codecs
 import getpass
 import twitter
 import config
@@ -38,6 +40,9 @@ class Commands:
 
 	def __init__(self):
 
+		if not sys.stdout.isatty():
+			sys.stdout = codecs.getwriter("UTF-8")(sys.stdout)
+
 		self.commands = {}
 		self.parser = OptionParser(usage="%prog COMMAND [options]",
 				version=VERSION)
@@ -81,11 +86,22 @@ class Commands:
 		command.do_command(options, args, self)
 
 		self.config.dirty_writeback()
+	
+	def output(self, str):
+		enc = sys.stdout.encoding
+		if enc is None:
+			print unicode(str)
+		else:
+			print unicode(str).encode(enc, "replace")
+
 
 class Command:
 	def __init__(self):
 		self.options = []
 
+	def get_options(self):
+		return self.options
+
 class DefaultCommand(Command):
 	def __str__(self):
 		return ""
@@ -107,15 +123,15 @@ class DefaultCommand(Command):
 	def do_command(self, options, args, base):
 		if options.commands:
 			base.parser.print_usage()
-			print "Available commands:"
+			base.output("Available commands:")
 			keys = base.commands.keys()
 			keys.sort()
 			for k in keys:
-				print " %10s %s" %\
-					(k, base.commands[k].get_help())
-			print ""
-			print "For command-specific help, use "\
-				"%sCOMMAND --help" % base.parser.get_prog_name()
+				base.output(" %12s %s" %\
+					(k, base.commands[k].get_help()))
+			base.output("")
+			base.output("For command-specific help, use "\
+				"%sCOMMAND --help" % base.parser.get_prog_name())
 		else:
 			base.parser.error("Command not found. Try --commands")
 
@@ -132,6 +148,9 @@ class CommandWithAuth:
 			help="Specifies PASS as your password.")
 		]
 
+	def get_options(self):
+		return self.options
+
         def passwdprompt(self, usr):
 		pwd = ""
 		for _ in range(3):
@@ -179,7 +198,7 @@ class TweetCommand(CommandWithAuth):
 	def get_options(self):
 		self.options.append(make_option("-r", "--replyto",
 			metavar="ID", default=-1, type="int",
-			help="Explicitly sets the ID of the status message to"
+			help="Explicitly sets the ID of the status message to "
 			"reply to."))
 		return self.options
 
@@ -189,21 +208,52 @@ class TweetCommand(CommandWithAuth):
 	def get_help(self):
 		return "Updates the authenticating user's Twitter status"
 
-	def do_command(self, options, args, base):
-		if len(args) < 1:
-			base.parser.error("No message specified.")
-
-		tweetmsg = " ".join(args)
+	def _do_tweet(self, options, base, msg):
+		if not msg:
+			return False
 
 		user, pwd = self.get_auth(options, base)
 		answer = self.callwithauth(base.twitter.status_update, options,
-				[tweetmsg, options.replyto], base, user, pwd)
+				[msg, options.replyto], base, user, pwd)
 
 		result = data.Status()
 		result.load_json(answer)
 
-		print unicode(result)
+		base.output(result)
+		return True
+	
+	def _get_tweets(self, args, base):
+		if not sys.stdin.isatty():
+			line = sys.stdin.readline()
+			while line:
+				if line.endswith("\n"):
+					line = line[:-1]
+				if line:
+					yield line
+				line = sys.stdin.readline()
+		elif len(args) > 0:
+			yield " ".join(args)
+		else:
+			base.parser.error("No message specified.")
 
+	def do_command(self, options, args, base):
+		for tweet in self._get_tweets(args, base):
+			self._do_tweet(options, base, tweet)
+
+
+class SingCommand(TweetCommand):
+	def __str__(self):
+		return "sing"
+
+	def get_help(self):
+		return "Similar to 'tweet', wraps the status in musical notes"
+
+	def do_command(self, options, args, base):
+		note = '\xe2\x99\xab'
+		for tweet in TweetCommand._get_tweets(self, args, base):
+			TweetCommand._do_tweet(self, options, base,
+					" ".join([note, tweet, note]))
+		
 
 class DirectCommand(CommandWithAuth):
 	def __str__(self):
@@ -235,9 +285,10 @@ class DirectCommand(CommandWithAuth):
 		result.load_json(answer)
 		
 		if not result.id:
-			raise TwitterException("Sending direct message failed.")
+			raise twitter.TwitterException(
+					"Sending direct message	failed.")
 
-		print unicode(result)
+		base.output(result)
 
 
 class DirectTLCommand(CommandWithAuth):
@@ -280,7 +331,7 @@ class DirectTLCommand(CommandWithAuth):
 		results = data.DirectList(answer)
 		
 		for result in results:
-			print unicode(result)
+			base.output(result)
 
 
 class DirectSentCommand(CommandWithAuth):
@@ -323,7 +374,7 @@ class DirectSentCommand(CommandWithAuth):
 		results = data.DirectList(answer)
 		
 		for result in results:
-			print unicode(result)
+			base.output(result)
 
 
 class DirectDelCommand(CommandWithAuth):
@@ -358,9 +409,9 @@ class DirectDelCommand(CommandWithAuth):
 			result = data.DirectMsg()
 			result.load_json(answer)
 
-			print u'Deleted: ' + unicode(result)
+			base.output(u'Deleted: ' + unicode(result))
 		else:
-			print "No messages deleted. Are you authorized?"
+			base.output("No messages deleted. Are you authorized?")
 	
 
 class PublicTLCommand(Command):
@@ -380,13 +431,17 @@ class PublicTLCommand(Command):
 	def do_command(self, options, args, base):
 
 		answer = base.twitter.status_public_timeline()
-		results = data.StatusList(answer)
+
+		try:
+			results = data.StatusList(answer)
+		except Exception, e:
+			raise twitter.TwitterException(e)
 
 		for result in results:
-			print unicode(result)
+			base.output(result)
 
 		if len(results) == 0:
-			print "(Nothing new)"
+			base.output("(Nothing new)")
 
 
 class FriendsTLCommand(CommandWithAuth):
@@ -401,6 +456,12 @@ class FriendsTLCommand(CommandWithAuth):
 		self.options.append(make_option("-P", "--page",
 			metavar="PAGE", type="int",
 			help="Lists the PAGEth page of status updates "))
+		self.options.append(make_option("-n", "--count",
+			metavar="COUNT",
+			help="The number of statuses to retrieve (max 200)"))
+		self.options.append(make_option("-q", "--quiet",
+			default=False, action="store_true",
+			help="Don't print the 'Nothing new' message."))
 		return self.options
 
 	def get_usage(self):
@@ -428,21 +489,26 @@ class FriendsTLCommand(CommandWithAuth):
 			catchup = True
 
 		answer = self.callwithauth(base.twitter.status_friends_timeline,
-				options, [since, since_id, options.page], base,
-				user, pwd)
+				options,
+				[since, since_id, options.page, options.count],
+				base, user, pwd)
 
-		results = data.StatusList(answer)
+		try:
+			results = data.StatusList(answer)
+		except Exception, e:
+			raise twitter.TwitterException(e)
 
 		if since:
-			print "Since %s:" % since
+			base.output("Since %s:" % since)
 		for result in results:
-			print unicode(result)
+			base.config.name_cache_add(result.user.screen_name)
+			base.output(result)
 
 		if len(results) > 0 and catchup:
 			base.config.set_stamp('friendstl', user,
 					results.get_last_id())
-		elif catchup:
-			print "(Nothing new)"
+		elif catchup and not options.quiet:
+			base.output("(Nothing new)")
 
 
 class UserTLCommand(CommandWithAuth):
@@ -488,10 +554,13 @@ class UserTLCommand(CommandWithAuth):
 		answer = self.callwithauth(base.twitter.status_user_timeline,
 				options, [ident, count, since, since_id,
 					options.page], base, user, pwd)
-		results = data.StatusList(answer)
+		try:
+			results = data.StatusList(answer)
+		except Exception, e:
+			raise twitter.TwitterException(e)
 
 		for result in results:
-			print unicode(result)
+			base.output(result)
 
 
 class ShowCommand(CommandWithAuth):
@@ -522,7 +591,7 @@ class ShowCommand(CommandWithAuth):
 			result = data.Status()
 			result.load_json(answer)
 			
-			print unicode(result)
+			base.output(result)
 
 
 class RepliesCommand(CommandWithAuth):
@@ -560,10 +629,13 @@ class RepliesCommand(CommandWithAuth):
 				pwd)
 
 		if answer:
-			results = data.StatusList(answer)
+			try:
+				results = data.StatusList(answer)
+			except Exception, e:
+				raise twitter.TwitterException(e)
 
 			for result in results:
-				 print unicode(result)
+				 base.output(result)
 
 
 class DeleteCommand(CommandWithAuth):
@@ -598,9 +670,9 @@ class DeleteCommand(CommandWithAuth):
 			result = data.Status()
 			result.load_json(answer)
 
-			print u'Deleted: ' + unicode(result)
+			base.output(u'Deleted: ' + unicode(result))
 		else:
-			print u'No messages deleted. Are you authorised?'
+			base.output('No messages deleted. Are you authorised?')
 
 
 class BlockCommand(CommandWithAuth):
@@ -628,7 +700,7 @@ class BlockCommand(CommandWithAuth):
 		answer = self.callwithauth(base.twitter.block_create, options,
 				[ident], base, user, pwd)
 		if answer:
-			print "User ID %s blocked." % str(ident)
+			base.output("User ID %s blocked." % str(ident))
 
 
 class UnblockCommand(CommandWithAuth):
@@ -656,7 +728,7 @@ class UnblockCommand(CommandWithAuth):
 		answer = self.callwithauth(base.twitter.block_destroy, options,
 				[ident], base, user, pwd)
 		if answer:
-			print "User ID %s unblocked." % str(ident)
+			base.output("User ID %s unblocked." % str(ident))
 
 class UserCommand(Command):
 	def __str__(self):
@@ -694,7 +766,7 @@ class UserCommand(Command):
 				p = "no password"
 			if i == default:
 				p += ", default"
-			print "%15s (%s)" % (i, p)
+			print("%15s (%s)" % (i, p))
 
 	def __set_options(self, options, base):
 		if options.password and options.user:
@@ -742,3 +814,79 @@ class UserCommand(Command):
 		else:
 			base.parser.error("Insufficient options specified")
 
+
+class NameCacheCommand(Command):
+	def __str__(self):
+		return "namecache"
+
+	def get_options(self):
+		self.options.append(make_option("-c", "--clear",
+			default=False, action="store_true",
+			help="Clear the name cache."))
+		self.options.append(make_option("-@", "--at",
+			default=False, action="store_true",
+			help="Prepend usernames with '@'."))
+		return self.options
+
+	def get_usage(self):
+		return "%prog [options]"
+
+	def get_help(self):
+		return "Access and manipulate the username cache."
+
+	def do_command(self, options, args, base):
+		if options.clear:
+			base.config.name_cache_clear()
+			return
+
+		if options.at:
+			for n in base.config.name_cache_list():
+				base.output("@%s" % n)
+		else:
+			for n in base.config.name_cache_list():
+				base.output(n)
+
+class IPLimitCommand(Command):
+	def __str__(self):
+		return "iplimit"
+
+	def get_usage(self):
+		return "%prog [options]"
+
+	def get_help(self):
+		return "Show the API rate limit for your IP address."
+
+	def do_command(self, options, args, base):
+
+		answer = base.twitter.account_rate_limit_status(False)
+
+		if answer:
+			try:
+				rl = data.RateLimit(answer)
+				base.output(str(rl)[:-1])
+			except Exception, e:
+				base.parser.error(e)
+
+
+class AccountLimitCommand(CommandWithAuth):
+	def __str__(self):
+		return "accountlimit"
+
+	def get_usage(self):
+		return "%prog [options]"
+
+	def get_help(self):
+		return "Show the API rate limit for your Twitter account."
+
+	def do_command(self, options, args, base):
+		user, pwd = self.get_auth(options, base)
+		answer = self.callwithauth(base.twitter.account_rate_limit_status,
+				options, [True], base, user, pwd)
+
+		if answer:
+			try:
+				rl = data.RateLimit(answer)
+				base.output(str(rl)[:-1])
+			except Exception, e:
+				base.parser.error(e)
+
